function[Lopt, Ropt, Zopt, infos, options, pred_testopt] = fixedrank_onesided_inductive_matrixcompletion(problem, Uinit, options)
%
%
% The function handles the side information on the rows.
% We solve for the loss ||A*W - Y|| in matrix completion with side information.
% A is the left side informaiton matrix.
%
% Wopt = Lopt * Ropt
%

    % Problem description
    d = problem.d;
    T = problem.T;
    r = problem.r;
    C = problem.C;
    I = problem.I; % Row indices
    J = problem.J; % Column indices
    Y = problem.Y; % Known entries
    
    A = problem.A; % Left side hand side information matrix.
    
    
    % sorting
    mask = sparse(double(I), double(J), Y, d, T, length(I));
    [I, J, Y] = find(mask);
    I = uint32(I);
    J = uint32(J);
    
    length_entries = length(I); % Number of known entries
    if isfield(problem,'I_test') && ~isempty(problem.I_test)
        I_test = problem.I_test;
        J_test = problem.J_test;
        Y_test = problem.Y_test;
        length_entries_test = length(I_test);
        A_test = problem.A_test; % Left side information matrix for the test data.
    end
    
    
    % Some sparse structures that are often used.
    mask = sparse(double(I), double(J), ones(length(I), 1), d, T, length(I));
    Z_sparse_mat = sparse(double(I), double(J), ones(length(I), 1), d, T, length(I));
    
    % Local defaults for options
    localdefaults.maxiter = 50; % Max iterations.
    localdefaults.tolgradnorm = 1e-6; % Absolute tolerance on Gradnorm.
    localdefaults.maxinner = 30; % Max inner iterations for the tCG step.
    localdefaults.tolrelgradnorm = 1e-10; % Gradnorm/initGradnorm tolerance.
    localdefaults.method = 'TR'; % Default solver is trustregions (TR).
    localdefaults.verbosity = 2; % Default: show the output.
    
    % Manifold geometry
    d1 = size(A, 2);
    problem.M = spectrahedronfactory(d1, r);
    
    
    % Initialization
    if ~exist('Uinit', 'var')
        Uinit = [];
    end
    
    if ~exist('options', 'var') || isempty(options)
        options = struct();
    end
    options = mergeOptions(localdefaults, options);
    
    
    % Cost
    problem.cost = @cost;
    function [f, store] = cost(U, store) % U translates to AU.
        if ~isfield(store, 'Z')
            %             [store.Z, store.UtZ, store.Zchol] = squareLoss_dual_get_Z(U,C,I, J, Y,mask);
            AU = full(A*U);
            [store.Z, store.UtAtZ, store.Zchol] = squareLoss_dual_get_Z(AU, C, I, J, Y, mask);
            store.AU = AU;
            
            %             size(A)
            %             size(AU)
            %             size(store.UtAtZ)
            %             max(I)
            %             size(mask)
            %             pause;
        end
        Z = store.Z;
        UtAtZ = store.UtAtZ;
        f = (Y'*Z) - (Z'*Z)/(4*C) - 0.5*norm(UtAtZ,'fro')^2;
    end
    
    % Gradient
    problem.egrad = @egrad;
    function [grad, store] = egrad(U, store) % U translates to AU.
        if ~isfield(store, 'Z')
            [~, store] = cost(U, store);
        end
        Z = store.Z;
        UtAtZ = store.UtAtZ;
        grad = -A'*mymultsparsefull(Z, UtAtZ', mask); % Equivalent to grad = -At*(Z*ZtAU).
        
    end
    
    % Hessian
    problem.ehess = @ehess;
    function [graddot, store] = ehess(U, Udot,store) % U translates to AU.
        if ~isfield(store, 'Z')
            [~, store] = cost(U, store);
        end
        Z = store.Z;
        UtAtZ = store.UtAtZ;
        Zchol = store.Zchol;
        AU = store.AU;
        
        % Get Zdot at U and Udot.
        %         [Zdot, UtZdot, UdottZ] = squareLoss_dual_get_Zdot(U,C,I, J,mask,Udot,Z,UtZ,Zchol);
        [Zdot, UtAtZdot, UdottAtZ] = squareLoss_dual_get_Zdot(AU,C,I, J,mask,A*Udot,Z,UtAtZ,Zchol);
        
        graddot = -mymultsparsefull(Zdot, UtAtZ', mask)...
            -mymultsparsefull(Z, (UtAtZdot+UdottAtZ)', mask);
        graddot = A'*graddot;
    end
    
    % Stats that we compute every iteration
    options.statsfun = @mystatsfun;
    function stats = mystatsfun(problem, U, stats, store) % U translates to AU.
        AU = store.AU;
        
        % Primal low-rank factors
        L = AU;
        R = store.UtAtZ;
        
        % Compute train RMSE
        error_train = myspmaskmult(L', R, I, J) ...
            - Y;
        stats.loss = norm(error_train)^2; % Square error on train data.
        stats.loss_rmse = sqrt((error_train'*error_train)/length_entries); % RMSE on train data.
        
        if exist('I_test', 'var')
            
            pred_test = myspmaskmult((A_test*U)', R, I_test, J_test) ; % Prediction function
            error_test = pred_test ...
                - Y_test;
            stats.loss_test = norm(error_test)^2; % Square error on test data.
            stats.loss_test_rmse = sqrt((error_test'*error_test)/length_entries_test); % RMSE on test data.
            stats.test_auc = mycomputeAUC(pred_test, Y_test);
            if options.verbosity >1
                fprintf('LossRMSE %e TestLossRMSE %e TestLossAUC %e\n', stats.loss_rmse, stats.loss_test_rmse, stats.test_auc);
            end
        end
    end
    
    % Additional stopping criteria
    options.stopfun = @mystopfun;
    function stopnow = mystopfun(problem, U, info, last)
        stopnow = (last >= 3 && info(last).gradnorm/info(1).gradnorm < options.tolrelgradnorm);
    end
    
    
    %     % Gradient and Hessian checks
    %     checkgradient(problem);
    %     pause;
    %     checkhessian(problem);
    %     pause;
    
    % Solver
    if strcmpi('TR', options.method)
        % Riemannian trustregions
        [Uopt,~,infos] = trustregions(problem, Uinit, options);
    elseif strcmpi('CG', options.method)
        % Riemannian conjugategradients
        options.beta_type = 'H-S';
        options.linesearch = @linesearch;
        options.ls_contraction_factor = .2;
        options.ls_optimism = 1.1;
        options.ls_suff_decr = 1e-4;
        options.ls_max_steps = 25;
        [Uopt,~,infos] = conjugategradient(problem, Uinit, options);
    end
    
    % Compute optimal Zopt
    [Zopt,UopttAtZ] = squareLoss_dual_get_Z(full(A*Uopt),C,I, J, Y,mask);
    
    % Optimal low-rank factor
    Lopt = Uopt;
    Ropt = UopttAtZ;

    if exist('I_test', 'var')
            pred_testopt = myspmaskmult(full((A_test*Lopt)'), Ropt, I_test, J_test) ;
    else
            pred_testopt = [];
    end
end
