function[Lopt, Ropt, Zopt, infos, options, pred_testopt] = fixedrank_twosided_inductive_matrixcompletion(problem, Uinit, options)
%
%
% The function handles both the side information on the rows and coloumns.
% We solve for the loss ||A*W*B - Y|| in matrix completion with side information.
% A and B are left and right side information matrices.
%
% Wopt = Lopt * Ropt.
% 

    % Problem description
    d = problem.d;
    T = problem.T;
    r = problem.r;
    C = problem.C;
    I = problem.I; % Row indices
    J = problem.J; % Column indices
    Y = problem.Y; % Known entries
    
    A = problem.A; % Left side hand side information matrix.
    B = problem.B; % Right side hand side ifnormation matrix.
    
    
    % sorting
    mask = sparse(double(I), double(J), Y, d, T, length(I));
    [I, J, Y] = find(mask);
    I = uint32(I);
    J = uint32(J);
    
    length_entries = length(I); % Number of known entries
    if isfield(problem,'I_test') && ~isempty(problem.I_test)
        I_test = problem.I_test;
        J_test = problem.J_test;
        Y_test = problem.Y_test;
        length_entries_test = length(I_test);
        A_test = problem.A_test; % Left side information matrix for the test data.
        B_test = problem.B_test; % Right side information matrix for the test data.
    end
    
    
    % Some sparse structures that are often used.
    mask = sparse(double(I), double(J), ones(length(I), 1), d, T, length(I));
    Z_sparse_mat = sparse(double(I), double(J), ones(length(I), 1), d, T, length(I));
    
    % Local defaults for options
    localdefaults.maxiter = 50; % Max iterations.
    localdefaults.tolgradnorm = 1e-6; % Absolute tolerance on Gradnorm.
    localdefaults.maxinner = 30; % Max inner iterations for the tCG step.
    localdefaults.tolrelgradnorm = 1e-10; % Gradnorm/initGradnorm tolerance.
    localdefaults.method = 'TR'; % Default solver is trustregions (TR).
    localdefaults.verbosity = 2; % Default: show the output.
    localdefaults.innerpcgtol = 1e-8; % Default: Inner PCG call for solving the linear system.
    localdefaults.innerpcgmaxiter = 2000; % Default: Inner PCG call for solving the linear system.
    
    % Manifold geometry
    d1 = size(A, 2);
    d2 = size(B, 2);
    
    problem.M = spectrahedronfactory(d1, r);
    
    
    % Initialization
    if ~exist('Uinit', 'var')
        Uinit = [];
    end
    
    if ~exist('options', 'var') || isempty(options)
        options = struct();
    end
    options = mergeOptions(localdefaults, options);
    
    
    
    % Cost
    problem.cost = @cost;
    function [f, store] = cost(U, store) % U translates to AU.
        if ~isfield(store, 'Z')
            AU = full(A*U);
            [store.Z, store.UtAtZ] = twosided_squareLoss_dual_get_Z(AU, C, I, J, Y, mask, B, options);
            store.AU = AU;
        end
        Z = store.Z;
        UtAtZ = store.UtAtZ;
        UtAtZBt = UtAtZ*B';
        f = (Y'*Z) - (Z'*Z)/(4*C) - 0.5*norm(UtAtZBt,'fro')^2;
    end
    
    % Gradient
    problem.egrad = @egrad;
    function [grad, store] = egrad(U, store) % U translates to AU.
        if ~isfield(store, 'Z')
            [~, store] = cost(U, store);
        end
        Z = store.Z;
        AtZ = mymultfullsparse(A',Z,mask);
        grad =  - AtZ*(B'*(B*(AtZ'*U)));% -A'*ZBt*((ZBt)'*AU); % Equivalent to grad = -At*(Z*Bt*B*ZtAU).
    end
    
    
    % Hessian
    problem.ehess = @ehess;
    function [graddot, store] = ehess(U, Udot,store) % U translates to AU.
        if ~isfield(store, 'Z')
            [~, store] = cost(U, store);
        end
        Z = store.Z;
        AU = store.AU;
        AUdot = A*Udot;
        
        % rhsdot
        AUdottZ = mymultfullsparse(AUdot', Z, mask); % |Omega| operations. Mex code heavy.
        AUdottZBt = AUdottZ*B';
        AUdottZBtB = full(AUdottZBt*B);
        
        AUtZ = mymultfullsparse(AU', Z, mask); % |Omega| operations. Mex code heavy.
        AUtZBt = AUtZ*B';
        AUtZBtB = full(AUtZBt*B);
        
        rhsdot = - myspmaskmult([AU AUdot]', [AUdottZBtB ;AUtZBtB ], I, J);
        
        % Get Zdot at U and Udot.
        [Zdot, UtAtZdot] = twosided_squareLoss_dual_get_Z(AU, C, I, J, rhsdot, mask, B, options);
        
        AtZ = mymultfullsparse(A',Z,mask);
        AtZdot = mymultfullsparse(A',Zdot,mask);
        
        graddot =  - AtZdot*(B'*(B*(AtZ'*U))) ...
            - AtZ*(B'*(B*(AtZdot'*U)))...
            - AtZ*(B'*(B*(AtZ'*Udot)));
    end
    
    
    
    % Stats that we compute every iteration
    options.statsfun = @mystatsfun;
    function stats = mystatsfun(problem, U, stats, store) % U translates to AU.
        % Primal low-rank factors
        L = U;
        R = store.UtAtZ*B';
        
        % Compute train RMSE
        error_train = myspmaskmult(full((A*L)'), full(R*B), I, J) ...
            - Y;
        stats.loss = norm(error_train)^2; % Square error on train data.
        stats.loss_rmse = sqrt((error_train'*error_train)/length_entries); % RMSE on train data.
        
        if exist('I_test', 'var')
            
            pred_test = myspmaskmult(full((A_test*U)'), full(R*B_test), I_test, J_test) ; % Prediction function
            error_test = pred_test ...
                - Y_test;
            stats.loss_test = norm(error_test)^2; % Square error on test data.
            stats.loss_test_rmse = sqrt((error_test'*error_test)/length_entries_test); % RMSE on test data.
            stats.test_auc = mycomputeAUC(pred_test, Y_test);
            if options.verbosity >1
                fprintf('LossRMSE %e TestLossRMSE %e TestLossAUC %e\n', stats.loss_rmse, stats.loss_test_rmse, stats.test_auc);
            end
        end
    end
    
    % Additional stopping criteria
    options.stopfun = @mystopfun;
    function stopnow = mystopfun(problem, U, info, last)
        stopnow = (last >= 3 && info(last).gradnorm/info(1).gradnorm < options.tolrelgradnorm);
    end
    
    
    %     % Gradient and Hessian checks
    %     checkgradient(problem);
    %     pause;
    %     checkhessian(problem);
    %     pause;
    
    % Solver
    if strcmpi('TR', options.method)
        % Riemannian trustregions
        [Uopt,~,infos] = trustregions(problem, Uinit, options);
    elseif strcmpi('CG', options.method)
        % Riemannian conjugategradients
        options.beta_type = 'H-S';
        options.linesearch = @linesearch;
        options.ls_contraction_factor = .2;
        options.ls_optimism = 1.1;
        options.ls_suff_decr = 1e-4;
        options.ls_max_steps = 25;
        [Uopt,~,infos] = conjugategradient(problem, Uinit, options);
    end
    
    % Compute optimal Zopt
    [Zopt,UopttAtZ] = twosided_squareLoss_dual_get_Z(full(A*Uopt), C, I, J, Y, mask, B, options);
    
    % Optimal low-rank factor
    Lopt = Uopt;
    Ropt = UopttAtZ*B';
    if exist('I_test', 'var')
    pred_testopt = myspmaskmult(full((A_test*Lopt)'), full(Ropt*B_test), I_test, J_test) ;
    else
        pred_testopt = [];
    end
    
end
